//
// $Id: halt_cpu.cc,v 1.3 2003/10/08 20:39:35 nj Exp $
//

#include "halt_cpu.h"

//
// FSM for halting cpu
//
void halt_cpu::execute()
{
  cpuwait_i.write(SC_LOGIC_1);
  pcwe.write(SC_LOGIC_0);
  regre.write(SC_LOGIC_0);
  regwe.write(SC_LOGIC_0);

  // register address when reading out registers...
  // sc_out<sc_uint<5> > regadr
  if (!reset.read())
    {
      cpuwait_i.write(SC_LOGIC_0);
      state = idle;
    }
  else
    {
      cout << "halt_cpu::execute()" << endl;
      cout << "state = " << state.read() << endl;
      cout << "regno = " << regno.read() << endl;

      switch(state)
	{
	case idle:
	  cpuwait_i.write(SC_LOGIC_0);
	  if (halt.read())
	    {
	      cpuwait_i.write(SC_LOGIC_1);
	      savepc = pc_in.read();
	      state = start_state;
	    }
	  break;
	case start_state:
	  regno = 0;
	  if (halt.read())
	    state = copyregs;
	  else
	    state = idle;
	  break;
	case copyregs:
	  regre.write(SC_LOGIC_1);
	  regadr = (sc_uint<5>)regno;
	  regno = regno + 1; // NB! signal
	  
	  if (regno == 32)
	    state = halted;
	  else
	    state = copyregs;
	  break;
	case halted:
	  // Missing: magic signal to turn V_DD off!!
	  
	  if (!halt.read())
	    {
	      regno = 0;
	      state = copyback;
	    }
	  break;
	case copyback:
	  regwe.write(SC_LOGIC_1);
	  regadr = (sc_uint<5>)regno;
	  regno = regno + 1; // NB! signal
	  
	  if (regno == 32)
	    state = end_state;
	  else
	    state = copyback;
	  break;
	case end_state:
	  pcwe = SC_LOGIC_1;
	  pc_out = savepc;
	  state = idle;
	  break;
	}
    }
}

//
// Save the registers on falling clock edge
//
void halt_cpu::execute_neg()
{
  if ((state.read() == copyregs) && (regno < 32))
    {
      saveregs[regno.read()] = reg_in.read();
    }
  else if ((state.read() == copyback) && (regno < 32))
    {
      reg_out = saveregs[regno.read()];
    }
}

//
// Set signal...immediately
//
void halt_cpu::setcpuwait()
{
  if (halt.read())
    cpuwait = cpuwait_i | SC_LOGIC_1;
  else
    cpuwait = cpuwait_i | SC_LOGIC_0;
}
