//
// $Id: pm_fetch_dec.h,v 1.21 2003/10/17 09:35:14 nj Exp $
//
#ifndef _PM_FETCH_DEC_H
#define _PM_FETCH_DEC_H
#include <systemc.h>
#include "config.h"
#include "constants.h"

#include "avrucpackage.h"

#define DBUS_MUX_GENERATE(x) v_dbusout_int[x] = (v_reg_rd_out[x] & (/*idc_push*/push_st | \
  (idc_st_x | idc_st_y | idc_std_y | idc_st_z | idc_std_z))) | \
  (v_gp_reg_tmp[x] & (st_st | sts_st1 | sts_st2)) | \
  (v_bitpr_io_out[x] & (cbi_st | sbi_st)) | \
  (v_program_counter_low_fr[x] & (/* idc_rcall |idc_icall | call_st1 */ rcall_st2 | icall_st2 | call_st3 )) | \
  (v_program_counter_high_fr[x] & (rcall_st1 | icall_st1 | call_st2)) | \
  (v_pc_for_interrupt[x] & irq_st1) | (v_pc_for_interrupt[x+8] & irq_st2) | \
  (v_reg_rd_out[x] & idc_out);

SC_MODULE(pm_fetch_dec)
{
  sc_in<bool> clk;
  sc_in<bool> nrst;

  sc_in<sc_logic> cpuwait;

  // Program memory ports
  sc_out<sc_uint<16> > pc;
  sc_in<sc_lv<16> > inst;

  // I/O Register ports
  sc_out<sc_uint<6> > adr;
  sc_out<sc_logic> iore;
  sc_out<sc_logic> iowe;

  // Data memory ports
  sc_out<sc_uint<16> > ramadr;
  sc_out<sc_logic> ramre;
  sc_out<sc_logic> ramwe;
  sc_in<sc_lv<8> > dbusin;
  sc_out<sc_lv<8> > dbusout;

  // Interrupt ports
  sc_in<sc_lv<23> > irqlines;
  sc_out<sc_logic> irqack;
  sc_out<sc_lv<5> > irqackad;

  // Interfaces to the alu
  sc_out<sc_lv<8> > alu_data_r_in;
  sc_out<sc_lv<8> > alu_data_d_in;

  // Operation signals inputs
  sc_out<sc_logic> idc_add_out;
  sc_out<sc_logic> idc_adc_out;
  sc_out<sc_logic> idc_adiw_out;
  sc_out<sc_logic> idc_sub_out;
  sc_out<sc_logic> idc_subi_out;
  sc_out<sc_logic> idc_sbc_out;
  sc_out<sc_logic> idc_sbci_out;
  sc_out<sc_logic> idc_sbiw_out;
  sc_out<sc_logic> adiw_st_out;
  sc_out<sc_logic> sbiw_st_out;
  sc_out<sc_logic> idc_and_out;
  sc_out<sc_logic> idc_andi_out;
  sc_out<sc_logic> idc_or_out;
  sc_out<sc_logic> idc_ori_out;
  sc_out<sc_logic> idc_eor_out;
  sc_out<sc_logic> idc_com_out;
  sc_out<sc_logic> idc_neg_out;
  sc_out<sc_logic> idc_inc_out;
  sc_out<sc_logic> idc_dec_out;
  sc_out<sc_logic> idc_cp_out;
  sc_out<sc_logic> idc_cpc_out;
  sc_out<sc_logic> idc_cpi_out;
  sc_out<sc_logic> idc_cpse_out;
  sc_out<sc_logic> idc_lsr_out;
  sc_out<sc_logic> idc_ror_out;
  sc_out<sc_logic> idc_asr_out;
  sc_out<sc_logic> idc_swap_out;

  // Alu data output
  sc_in<sc_lv<8> > alu_data_out;

  // Alu flags
  sc_in<sc_logic> alu_c_flag_out;
  sc_in<sc_logic> alu_z_flag_out;
  sc_in<sc_logic> alu_n_flag_out;
  sc_in<sc_logic> alu_v_flag_out;
  sc_in<sc_logic> alu_s_flag_out;
  sc_in<sc_logic> alu_h_flag_out;

  // Interfaces to the GPR
  sc_out<sc_lv<8> > reg_rd_in;
  sc_in<sc_lv<8> > reg_rd_out;
  sc_out<sc_lv<5> > reg_rd_adr;
  sc_in<sc_lv<8> > reg_rr_out;
  sc_out<sc_lv<5> > reg_rr_adr;
  sc_out<sc_logic> reg_rd_wr;
  sc_out<sc_logic> post_inc;
  sc_out<sc_logic> pre_dec;
  sc_out<sc_logic> reg_h_wr;
  sc_in<sc_lv<16> > reg_h_out;
  sc_out<sc_lv<3> > reg_h_adr;
  sc_in<sc_lv<16> > reg_z_out;

  // Interfaces to the input/output regfile
  sc_out<sc_lv<8> > sreg_fl_in;
  sc_in<sc_lv<8> > sreg_out;
  sc_out<sc_lv<8> > sreg_fl_wr_en;

  sc_in<sc_lv<8> > spl_out;
  sc_in<sc_lv<8> > sph_out;
  sc_out<sc_logic> sp_ndown_up;
  sc_out<sc_logic> sp_en;
  sc_in<sc_lv<8> > rampz_out;

  // Interfaces to the bit_processor
  sc_out<sc_lv<3> > bit_num_r_io;
  sc_in<sc_lv<8> > bitpr_io_out;
  sc_out<sc_lv<3> > branch;
  sc_in<sc_lv<8> > bit_pr_sreg_out;
  sc_out<sc_lv<3> > sreg_bit_num;
  sc_in<sc_lv<8> > bld_op_out;
  sc_in<sc_logic> bit_test_op_out;
  sc_out<sc_logic> idc_sbi_out;
  sc_out<sc_logic> sbi_st_out;
  sc_out<sc_logic> idc_cbi_out;
  sc_out<sc_logic> cbi_st_out;

  // Instructions and states
  sc_out<sc_logic> idc_bld_out;
  sc_out<sc_logic> idc_bst_out;
  sc_out<sc_logic> idc_bset_out;
  sc_out<sc_logic> idc_bclr_out;
  sc_out<sc_logic> idc_sbic_out;
  sc_out<sc_logic> idc_sbis_out;
  sc_out<sc_logic> idc_sbrs_out;
  sc_out<sc_logic> idc_sbrc_out;
  sc_out<sc_logic> idc_brbs_out;
  sc_out<sc_logic> idc_brbc_out;
  sc_out<sc_logic> idc_reti_out;

  // sleep instruction is being executed
  sc_out<sc_logic> sleepi;

  // signals for saving and restoring regs/pc
  sc_in<sc_logic> pcwe;
  sc_out<sc_uint<16> > pc_in;
  sc_in<sc_uint<16> > pc_out;


  // Signals
  sc_signal<sc_lv<16> > ramadr_reg_in;
  sc_signal<sc_logic> ramadr_reg_en;
  sc_signal<sc_logic> irqack_int;
  sc_signal<sc_lv<5> > irqackad_int;
  sc_signal<sc_logic> two_word_inst;
  sc_signal<sc_lv<16> > ram_adr_int;
  sc_signal<sc_lv<5> > adiw_sbiw_encoder_out;
  sc_signal<sc_lv<5> > adiw_sbiw_encoder_mux_out;

  sc_signal<sc_uint<16> > program_counter_tmp;
  sc_signal<sc_uint<16> > program_counter;
  sc_signal<sc_uint<16> > program_counter_in;
  sc_signal<sc_uint<8> > program_counter_high_fr;
  sc_signal<sc_uint<8> > program_counter_low_fr;
  sc_signal<sc_uint<8> > pc_low;
  sc_signal<sc_uint<8> > pc_high;

  sc_signal<sc_logic> pc_low_en;
  sc_signal<sc_logic> pc_high_en;
  sc_signal<sc_lv<16> > offset_brbx;
  sc_signal<sc_lv<16> > offset_rxx;
  sc_signal<sc_logic> pa15_pm;
  sc_signal<sc_logic> alu_reg_wr;
  sc_signal<sc_logic> reg_file_adr_space;
  sc_signal<sc_logic> io_file_adr_space;
  sc_signal<sc_logic> irq_start;

  // All *st signals are combinatorically calculated
  sc_signal<sc_logic> nirq_st0;
  sc_signal<sc_logic> irq_st1;
  sc_signal<sc_logic> irq_st2;
  sc_signal<sc_logic> irq_st3;
  sc_signal<sc_logic> ncall_st0;
  sc_signal<sc_logic> call_st1;
  sc_signal<sc_logic> call_st2;
  sc_signal<sc_logic> call_st3;
  sc_signal<sc_logic> nrcall_st0;
  sc_signal<sc_logic> rcall_st1;
  sc_signal<sc_logic> rcall_st2;
  sc_signal<sc_logic> nicall_st0;
  sc_signal<sc_logic> icall_st1;
  sc_signal<sc_logic> icall_st2;
  sc_signal<sc_logic> njmp_st0;
  sc_signal<sc_logic> jmp_st1;
  sc_signal<sc_logic> jmp_st2;
  sc_signal<sc_logic> ijmp_st;
  sc_signal<sc_logic> rjmp_st;
  sc_signal<sc_logic> nret_st0;
  sc_signal<sc_logic> ret_st1;
  sc_signal<sc_logic> ret_st2;
  sc_signal<sc_logic> ret_st3;
  sc_signal<sc_logic> nreti_st0;
  sc_signal<sc_logic> reti_st1;
  sc_signal<sc_logic> reti_st2;
  sc_signal<sc_logic> reti_st3;
  sc_signal<sc_logic> brxx_st;
  sc_signal<sc_logic> adiw_st;
  sc_signal<sc_logic> sbiw_st;
  sc_signal<sc_logic> nskip_inst_st0;
  sc_signal<sc_logic> skip_inst_st1;
  sc_signal<sc_logic> skip_inst_st2;
  sc_signal<sc_logic> skip_inst_start;
  sc_signal<sc_logic> nlpm_st0;
  sc_signal<sc_logic> lpm_st1;
  sc_signal<sc_logic> lpm_st2;
  sc_signal<sc_logic> nelpm_st0;
  sc_signal<sc_logic> elpm_st1;
  sc_signal<sc_logic> elpm_st2;
  sc_signal<sc_logic> nsts_st0;
  sc_signal<sc_logic> sts_st1;
  sc_signal<sc_logic> sts_st2;
  sc_signal<sc_logic> nlds_st0;
  sc_signal<sc_logic> lds_st1;
  sc_signal<sc_logic> lds_st2;
  sc_signal<sc_logic> st_st;
  sc_signal<sc_logic> ld_st;
  sc_signal<sc_logic> sbi_st;
  sc_signal<sc_logic> cbi_st;
  sc_signal<sc_logic> push_st;
  sc_signal<sc_logic> pop_st;
  sc_signal<sc_logic> nop_insert_st;

  // *int signals are optimized away in synthesis
  sc_signal<sc_logic> cpu_busy;
  sc_signal<sc_lv<16> > pc_int;
  sc_signal<sc_lv<6> > adr_int;
  sc_signal<sc_logic> iore_int;
  sc_signal<sc_logic> iowe_int;
  sc_signal<sc_lv<16> > ramadr_int;
  sc_signal<sc_logic> ramre_int;
  sc_signal<sc_logic> ramwe_int;
  sc_signal<sc_lv<8> > dbusout_int;
  sc_signal<sc_lv<16> > instruction_reg;
  sc_signal<sc_lv<16> > instruction_code_reg;
  sc_signal<sc_logic> instruction_reg_ena;
  sc_signal<sc_logic> irq_int;
  sc_signal<sc_lv<16> > irq_vector_adr;
  sc_signal<sc_lv<7> > irq_pr_encoder;
  sc_signal<sc_uint<16> > pc_for_interrupt;

  // dex* signals are optimized away in synthesis
  sc_signal<sc_lv<8> > dex_dat8_immed;
  sc_signal<sc_lv<6> > dex_dat6_immed;
  sc_signal<sc_lv<12> > dex_adr12mem_s;
  sc_signal<sc_lv<6> > dex_adr6port;
  sc_signal<sc_lv<5> > dex_adr5port;
  sc_signal<sc_lv<6> > dex_adr_disp;
  sc_signal<sc_lv<3> > dex_condition;
  sc_signal<sc_lv<3> > dex_bitnum_sreg;
  sc_signal<sc_lv<5> > dex_adrreg_r;
  sc_signal<sc_lv<5> > dex_adrreg_d;
  sc_signal<sc_lv<3> > dex_bitop_bitnum;
  sc_signal<sc_lv<7> > dex_brxx_offset;
  sc_signal<sc_lv<2> > dex_adiw_sbiw_reg_adr;
  sc_signal<sc_lv<5> > dex_adrreg_d_latched;
  sc_signal<sc_lv<8> > gp_reg_tmp;
  sc_signal<sc_lv<5> > cbi_sbi_io_adr_tmp;
  sc_signal<sc_lv<3> > cbi_sbi_bit_num_tmp;

  sc_signal<sc_logic> idc_adc;
  sc_signal<sc_logic> idc_add;
  sc_signal<sc_logic> idc_adiw;
  sc_signal<sc_logic> idc_and;
  sc_signal<sc_logic> idc_andi;
  sc_signal<sc_logic> idc_asr;
  sc_signal<sc_logic> idc_bclr;
  sc_signal<sc_logic> idc_bld;
  sc_signal<sc_logic> idc_brbc;
  sc_signal<sc_logic> idc_brbs;
  sc_signal<sc_logic> idc_bset;
  sc_signal<sc_logic> idc_bst;
  sc_signal<sc_logic> idc_call;
  sc_signal<sc_logic> idc_cbi;
  sc_signal<sc_logic> idc_com;
  sc_signal<sc_logic> idc_cp;
  sc_signal<sc_logic> idc_cpc;
  sc_signal<sc_logic> idc_cpi;
  sc_signal<sc_logic> idc_cpse;
  sc_signal<sc_logic> idc_dec;
  sc_signal<sc_logic> idc_elpm;
  sc_signal<sc_logic> idc_eor;
  sc_signal<sc_logic> idc_icall;
  sc_signal<sc_logic> idc_ijmp;
  sc_signal<sc_logic> idc_in;
  sc_signal<sc_logic> idc_inc;
  sc_signal<sc_logic> idc_jmp;
  sc_signal<sc_logic> idc_ld_x;
  sc_signal<sc_logic> idc_ld_y;
  sc_signal<sc_logic> idc_ldd_y;
  sc_signal<sc_logic> idc_ld_z;
  sc_signal<sc_logic> idc_ldd_z;
  sc_signal<sc_logic> idc_ldi;
  sc_signal<sc_logic> idc_lds;
  sc_signal<sc_logic> idc_lpm;
  sc_signal<sc_logic> idc_lsr;
  sc_signal<sc_logic> idc_mov;
  sc_signal<sc_logic> idc_mul;
  sc_signal<sc_logic> idc_neg;
  sc_signal<sc_logic> idc_nop;
  sc_signal<sc_logic> idc_or;
  sc_signal<sc_logic> idc_ori;
  sc_signal<sc_logic> idc_out;
  sc_signal<sc_logic> idc_pop;
  sc_signal<sc_logic> idc_push;
  sc_signal<sc_logic> idc_rcall;
  sc_signal<sc_logic> idc_ret;
  sc_signal<sc_logic> idc_reti;
  sc_signal<sc_logic> idc_rjmp;
  sc_signal<sc_logic> idc_ror;
  sc_signal<sc_logic> idc_sbc;
  sc_signal<sc_logic> idc_sbci;
  sc_signal<sc_logic> idc_sbi;
  sc_signal<sc_logic> idc_sbic;
  sc_signal<sc_logic> idc_sbis;
  sc_signal<sc_logic> idc_sbiw;
  sc_signal<sc_logic> idc_sbrc;
  sc_signal<sc_logic> idc_sbrs;
  sc_signal<sc_logic> idc_sleep;
  sc_signal<sc_logic> idc_st_x;
  sc_signal<sc_logic> idc_st_y;
  sc_signal<sc_logic> idc_std_y;
  sc_signal<sc_logic> idc_st_z;
  sc_signal<sc_logic> idc_std_z;
  sc_signal<sc_logic> idc_sts;
  sc_signal<sc_logic> idc_sub;
  sc_signal<sc_logic> idc_subi;
  sc_signal<sc_logic> idc_swap;
  sc_signal<sc_logic> idc_wdr;
  sc_signal<sc_logic> idc_psinc;
  sc_signal<sc_logic> idc_prdec;
  sc_signal<sc_lv<8> > sreg_bop_wr_en;
  sc_signal<sc_logic> sreg_adr_eq;

  sc_signal<sc_lv<5> > x_reg_rd_adr;

  void instruction_fetch();
  void latcht_rd_adr();
  void adiw_sbiw_high_reg_adr();
  void ramadr_reg();
  void gp_reg_adr();
  void io_reg_adr();
  void datamemoryread();
  void datamemorywrite();
  void gp_registers_trig();
  void program_counter_high_store();
  void program_counter_for_lpm_elpm();
  void program_counter_both();
  // void program_counter_low();
  // void program_counter_high();
  void program_counter_for_interrupt();
  void skip_instruction_sm();
  void alu_state_machines();
  void lpm_state_machine();
  void lds_state_machine();
  void sts_state_machine();
  void jmp_state_machine();
  void rcall_state_machine();
  void icall_state_machine();
  void call_state_machine();
  void ret_state_machine();
  void reti_state_machine();
  void irq_state_machine();
  void irqack_reg();
  void irqackad_reg();
  void rjmp_push_pop_ijmp_state_brxx_machine();
  void ld_st_state_machine();
  void sbi_cbi_machine();
  void two_word_inst_detect();
  void data_extractor();
  void set_adiw_sbiw_encoder_out();
  void nop_insert_sm();
  void set_nop_insert_st();
  void instruction_decoder();
  void incdec_signals();
  void ramadr_signals();
  void rd_enable_signals();
  void reg_rd_adr_register();
  void rd_input_signals();
  void iore_logic();
  void adr_bus_mux();
  void dbusout_mux();
  void alu_connection();
  void set_pa15_pm();
  void set_offsets();
  void set_pc_signals();
  void set_pc_signal();
  // void set_program_counter_signal();
  void set_skip_inst_start();
  void interrupt_logic();
  void irq_flags();
  void sreg_flags_en_logic();
  void output_everything();

  SC_CTOR(pm_fetch_dec)
  {
    SC_METHOD(instruction_fetch);
    sensitive_pos << clk;
    SC_METHOD(latcht_rd_adr);
    sensitive_pos << clk;
    SC_METHOD(adiw_sbiw_high_reg_adr);
    sensitive_pos << clk;
    SC_METHOD(ramadr_reg);
    sensitive_pos << clk;
    SC_METHOD(gp_reg_adr);
    sensitive_pos << clk;
    SC_METHOD(io_reg_adr);
    sensitive_pos << clk;
    SC_METHOD(datamemoryread);
    sensitive_pos << clk;
    SC_METHOD(datamemorywrite);
    sensitive_pos << clk;
    SC_METHOD(gp_registers_trig);
    sensitive_pos << clk;
    SC_METHOD(program_counter_high_store);
    sensitive_pos << clk;
    SC_METHOD(program_counter_for_lpm_elpm);
    sensitive_pos << clk;
    SC_METHOD(program_counter_both);
    sensitive_pos << clk;
    /*
    SC_METHOD(program_counter_low);
    sensitive_pos << clk;
    SC_METHOD(program_counter_high);
    sensitive_pos << clk;
    */
    SC_METHOD(program_counter_for_interrupt);
    sensitive_pos << clk;
    SC_METHOD(skip_instruction_sm);
    sensitive_pos << clk;

    SC_METHOD(alu_state_machines);
    sensitive_pos << clk;
    SC_METHOD(lpm_state_machine);
    sensitive_pos << clk;
    SC_METHOD(lds_state_machine);
    sensitive_pos << clk;
    SC_METHOD(sts_state_machine);
    sensitive_pos << clk;
    SC_METHOD(jmp_state_machine);
    sensitive_pos << clk;
    SC_METHOD(rcall_state_machine);
    sensitive_pos << clk;
    SC_METHOD(icall_state_machine);
    sensitive_pos << clk;
    SC_METHOD(call_state_machine);
    sensitive_pos << clk;
    SC_METHOD(ret_state_machine);
    sensitive_pos << clk;
    SC_METHOD(reti_state_machine);
    sensitive_pos << clk;
    SC_METHOD(irq_state_machine);
    sensitive_pos << clk;

    SC_METHOD(irqack_reg);
    sensitive_pos << clk;
    SC_METHOD(irqackad_reg);
    sensitive_pos << clk;
    SC_METHOD(rjmp_push_pop_ijmp_state_brxx_machine);
    sensitive_pos << clk;
    SC_METHOD(ld_st_state_machine);
    sensitive_pos << clk;
    SC_METHOD(sbi_cbi_machine);
    sensitive_pos << clk;    

    // Concurrent statements
    SC_METHOD(two_word_inst_detect);
    sensitive << instruction_reg;
    SC_METHOD(data_extractor);
    sensitive << instruction_reg;
    SC_METHOD(set_adiw_sbiw_encoder_out);
    sensitive << dex_adiw_sbiw_reg_adr;

    SC_METHOD(nop_insert_sm);
    sensitive << instruction_reg << nop_insert_st << instruction_code_reg;
    SC_METHOD(set_nop_insert_st);
    sensitive << adiw_st << sbiw_st << cbi_st << sbi_st << rjmp_st << ijmp_st << pop_st << push_st << brxx_st << ld_st << st_st << ncall_st0 << nirq_st0 << nret_st0 << nreti_st0 << nlpm_st0 << njmp_st0 << nrcall_st0 << nicall_st0 << nsts_st0 << nlds_st0 << nskip_inst_st0;
    SC_METHOD(instruction_decoder);
    sensitive << instruction_code_reg;
    SC_METHOD(incdec_signals);
    sensitive << instruction_code_reg << idc_st_x << idc_st_y << idc_st_z << idc_ld_x << idc_ld_y << idc_ld_z;

    SC_METHOD(ramadr_signals);
    sensitive << sph_out << spl_out << reg_h_out << dex_adr_disp;
    sensitive << idc_ld_x << idc_ld_y << idc_ldd_y << idc_ld_z << idc_ldd_z << lds_st1 << idc_st_x;
    sensitive << idc_st_y << idc_std_y << idc_st_z << idc_std_z << sts_st1 << idc_push << idc_pop;
    sensitive << idc_rcall << rcall_st1 << cpuwait << idc_icall << icall_st1 << call_st1 << call_st2;
    sensitive << irq_st1 << irq_st2 << idc_ret << ret_st1 << idc_reti << reti_st1;
    sensitive << instruction_reg;

    SC_METHOD(rd_enable_signals);
    sensitive << dex_adrreg_d << ramadr_int << sbiw_st << adiw_st << alu_reg_wr << x_reg_rd_adr;
    sensitive << adiw_sbiw_encoder_out << adiw_sbiw_encoder_mux_out << lpm_st2;
    sensitive << pop_st << dex_adrreg_d_latched << reg_file_adr_space << lds_st2;
    sensitive << ld_st << sts_st2 << st_st << dex_adrreg_r;
    sensitive << idc_adc << idc_add << idc_adiw << adiw_st << idc_sub << idc_subi << idc_sbc;
    sensitive << idc_sbci << idc_sbiw << sbiw_st << idc_and << idc_andi << idc_or << idc_ori << idc_eor;
    sensitive << idc_com << idc_neg << idc_inc << idc_dec << idc_lsr << idc_ror << idc_asr << idc_swap;
    sensitive << idc_mov << idc_bld << idc_in << idc_ldi << idc_cpi << push_st;

    SC_METHOD(reg_rd_adr_register);
    sensitive_pos << clk;

    SC_METHOD(rd_input_signals);
    sensitive << reg_z_out << instruction_reg;
    sensitive << idc_in << lds_st2 << idc_bld << idc_mov << idc_ldi << ld_st;
    sensitive << reg_file_adr_space << pop_st << dbusin << reg_rr_out << gp_reg_tmp;
    sensitive << bld_op_out << reg_rr_out << dex_dat8_immed << alu_data_out;
    sensitive << sts_st2 << st_st << lpm_st2;

    SC_METHOD(iore_logic);
    sensitive << idc_in << idc_sbi << idc_cbi << idc_sbic << idc_sbis << ld_st << lds_st2 << io_file_adr_space << idc_out << sbi_st << cbi_st << st_st << sts_st2;
    SC_METHOD(adr_bus_mux);
    sensitive << ramadr_int << idc_in << idc_out << dex_adr6port << idc_cbi << idc_sbi << idc_sbic << idc_sbis << dex_adr5port << cbi_st << sbi_st << cbi_sbi_io_adr_tmp;
    SC_METHOD(dbusout_mux);
    sensitive << reg_rd_out << gp_reg_tmp << bitpr_io_out << program_counter << program_counter_high_fr << program_counter_low_fr << pc_for_interrupt;
    sensitive << idc_push << idc_st_x << idc_st_y << idc_std_y << idc_st_z << idc_std_z;
    sensitive << st_st << sts_st1 << sts_st2 << cbi_st << sbi_st << idc_rcall << idc_icall;
    sensitive << call_st1 << rcall_st1 << icall_st1 << call_st2 << irq_st1 << irq_st2 << idc_out;
    sensitive << rcall_st2 << call_st3 << icall_st2 << push_st;

    SC_METHOD(alu_connection);
    sensitive << idc_subi << idc_sbci << idc_andi << idc_ori << idc_cpi << dex_dat8_immed << idc_adiw << idc_sbiw << dex_dat6_immed << adiw_st << sbiw_st << reg_rr_out << reg_rd_out;
    SC_METHOD(set_pa15_pm);
    sensitive << rampz_out << idc_elpm;
    SC_METHOD(set_offsets);
    sensitive << dex_brxx_offset << dex_adr12mem_s;

    SC_METHOD(set_pc_signals);
    sensitive << offset_brbx << offset_rxx;
    sensitive << pa15_pm << reg_z_out << irqackad_int << dbusin << program_counter;
    sensitive << bit_test_op_out;
    sensitive << jmp_st1 << call_st1 << irq_st1 << ret_st1 << reti_st1 << ret_st2 << reti_st2 << lpm_st1;
    sensitive << program_counter_tmp << sts_st1 << lds_st1 << rcall_st1 << icall_st1 << call_st2 << irq_st2 << cpuwait << ret_st1 << reti_st1;
    sensitive << idc_brbc << idc_brbs << idc_rjmp << idc_rcall << idc_lpm << idc_elpm << idc_ld_x << idc_ld_y << idc_ld_z << idc_ldd_y << idc_ldd_z << idc_st_x << idc_st_y << idc_st_z << idc_std_y << idc_std_z << idc_adiw << idc_sbiw << idc_push << idc_pop << idc_cbi << idc_sbi;
    sensitive << idc_ijmp << idc_icall << instruction_reg;

    /*
    SC_METHOD(set_program_counter_signal);
    sensitive << pc_high << pc_low;
    */

    SC_METHOD(set_pc_signal);
    sensitive << program_counter;
    SC_METHOD(set_skip_inst_start);
    sensitive << idc_sbrc << idc_sbrs << idc_sbic << idc_sbis << idc_cpse;
    sensitive << bit_test_op_out << alu_z_flag_out;

    SC_METHOD(interrupt_logic);
    sensitive << irqlines;

    SC_METHOD(irq_flags);
    sensitive << sreg_bop_wr_en << sreg_adr_eq << dbusout_int << sreg_out << adr_int;
    sensitive << irqack_int << irqackad_int;
    sensitive << idc_adiw << idc_sbiw << idc_cbi << idc_sbi << idc_rjmp ;

    SC_METHOD(sreg_flags_en_logic);
    sensitive << dex_bitnum_sreg << reti_st3 << alu_h_flag_out << alu_s_flag_out;
    sensitive << alu_v_flag_out << alu_n_flag_out << alu_z_flag_out << alu_c_flag_out;
    sensitive << sbiw_st << adiw_st << irq_st1 << bit_pr_sreg_out;
    sensitive << idc_add << idc_adc << idc_adiw << adiw_st << idc_sub  << idc_subi << idc_sbc << idc_sbci << idc_sbiw << sbiw_st << idc_cp << idc_cpc << idc_cpi << idc_and << idc_andi << idc_or << idc_ori << idc_eor << idc_com << idc_neg << idc_inc << idc_dec << idc_lsr << idc_ror << idc_asr;
    sensitive << idc_bset << idc_bclr << idc_bst;

    SC_METHOD(output_everything);
    sensitive << idc_adc << idc_add << idc_adiw << idc_and << idc_andi << idc_asr << idc_bclr;
    sensitive << idc_bld << idc_brbc << idc_brbs << idc_bset << idc_bst << idc_call << idc_cbi;
    sensitive << idc_com << idc_cp << idc_cpc << idc_cpi << idc_cpse << idc_dec << idc_elpm;
    sensitive << idc_eor << idc_icall << idc_ijmp << idc_in << idc_inc << idc_jmp << idc_ld_x;
    sensitive << idc_ld_y << idc_ldd_y << idc_ld_z << idc_ldd_z << idc_ldi << idc_lds << idc_lpm;
    sensitive << idc_lsr << idc_mov << idc_mul << idc_neg << idc_nop << idc_or << idc_ori << idc_out;
    sensitive << idc_pop << idc_push << idc_rcall << idc_ret << idc_reti << idc_rjmp << idc_ror;
    sensitive << idc_sbc << idc_sbci << idc_sbi << idc_sbic << idc_sbis << idc_sbiw << idc_sbrc;
    sensitive << idc_sbrs << idc_sleep << idc_st_x << idc_st_y << idc_std_y << idc_st_z << idc_std_z;
    sensitive << idc_sts << idc_sub << idc_subi << idc_swap << idc_wdr << idc_psinc << idc_prdec;
    sensitive << adiw_st << sbiw_st << irq_st2 << irq_st1 << call_st2 << call_st1;
    sensitive << reti_st1 << ret_st1 << icall_st1 << cpuwait << rcall_st1 << dex_condition;
    sensitive << cbi_sbi_bit_num_tmp << sbi_st << cbi_st << dex_bitop_bitnum << dex_bitnum_sreg;
    sensitive << push_st << pop_st;

  }
};

#endif
