//
// $Id: service_module.h,v 1.5 2003/07/30 16:45:31 nj Exp $
//
#ifndef _SERVICE_MODULE_H
#define _SERVICE_MODULE_H
#include <systemc.h>
#include "config.h"
#include "constants.h"

#include "avrucpackage.h"

#define EIFR_GENERATE(x) v_eifr[x + 4] = (( ~v_eifr[x+4]) & v_edge_det_out[x]) | (v_eifr[x + 4] & (~(((eifr_sel.read() & iowe.read()) & v_dbus_in[x + 4]) | (v_extint_irq_ack[x] & (~v_edge_det_out[x])))));

#define EDGE_DET_GENERATE(x) v_edge_det_out[x] = ((~v_ext_int_inreg[x+4] & v_ext_int_del[x]) & (v_eicr[2*x+1] & ~v_eicr[2*x])) | ((v_ext_int_inreg[x+4] & ~v_ext_int_del[x]) & (v_eicr[2*x+1] & v_eicr[2*x]));

#define EXTINT_GENERATE(x) if (v_eicr[2*x+1] == 1) \
    v_extint_irq[x+4] = (v_eifr[x+4] & v_eimsk[x+4]); \
  else if ((v_eicr[2*x+1] == 0) & (v_eicr[2*x] == 0)) \
    v_extint_irq[x+4] = ~v_ext_int_inreg[x+4] & v_eimsk[x+4]; \
  else \
    v_extint_irq[x+4] = 0;

#define OUTPUT_DATA_GENERATE(x) v_dbus_out[x] = (v_mcucr[x] & mcucr_sel) | (v_eimsk[x] & eimsk_sel) | (v_eifr[x] & eifr_sel) | (v_eicr[x] & eicr_sel) | (v_mcusr[x] & mcusr_sel) | (v_xdiv[x] & xdiv_sel);

SC_MODULE(service_module)
{
  sc_in<bool> cp2;
  sc_in<bool> ireset;

  sc_in<sc_uint<6> > adr;
  sc_in<sc_lv<8> >   dbus_in;
  sc_out<sc_lv<8> >  dbus_out;
  sc_in<sc_logic>    iore;
  sc_in<sc_logic>    iowe;
  sc_out<sc_logic>   out_en;
  sc_out<sc_logic>   sleep_en;
  sc_out<sc_logic>   esram_en;
  sc_out<sc_logic>   esram_ws;
  sc_out<sc_lv<8> >  extint_irq;
  sc_in<sc_lv<4> >   extint_irq_ack;
  sc_in<sc_lv<8> >   ext_int_in;

  sc_signal<sc_lv<8> > mcucr;
  sc_signal<sc_lv<8> > eimsk;
  sc_signal<sc_lv<8> > eifr;
  sc_signal<sc_lv<8> > eicr;
  sc_signal<sc_lv<8> > mcusr;
  sc_signal<sc_lv<8> > xdiv;
  sc_signal<sc_logic>  mcucr_sel;
  sc_signal<sc_logic>  eimsk_sel;
  sc_signal<sc_logic>  eifr_sel;
  sc_signal<sc_logic>  eicr_sel;
  sc_signal<sc_logic>  mcusr_sel;
  sc_signal<sc_logic>  xdiv_sel;
  sc_signal<sc_lv<4> > ext_int_del;
  sc_signal<sc_lv<4> > edge_det_out;
  sc_signal<sc_lv<8> > ext_int_resync;
  sc_signal<sc_lv<8> > ext_int_inreg;

  void externalintsynchronizer();
  void externalintreg();
  void mcucr_reg();
  void eimsk_reg();
  void eicr_reg();
  void mcusr_reg();
#ifdef CI_ML_XDIV
  void xdiv_reg()
#endif
  void eifr_reg();
  void set_selsignals();
  void edge_detector_logic();
  void high_irqs();
  void output_data_mux();

  SC_CTOR(service_module)
  {
    SC_METHOD(externalintsynchronizer);
    sensitive_pos << cp2;

    SC_METHOD(externalintreg);
    sensitive_pos << cp2;
    
    SC_METHOD(mcucr_reg);
    sensitive_pos << cp2;
  
    SC_METHOD(eimsk_reg);
    sensitive_pos << cp2;

    SC_METHOD(eicr_reg);
    sensitive_pos << cp2;

    SC_METHOD(mcusr_reg);
    sensitive_pos << cp2;

#ifdef CI_ML_XDIV
    SC_METHOD(xdiv_reg);
    sensitive_pos << cp2;
#endif

    SC_METHOD(eifr_reg);
    sensitive_pos << cp2;

    SC_METHOD(set_selsignals);
    sensitive << adr << mcucr_sel << eimsk_sel << eifr_sel << eicr_sel << mcusr_sel << xdiv_sel << iore;

    SC_METHOD(edge_detector_logic);
    sensitive << ext_int_inreg << ext_int_del << eicr;

    SC_METHOD(high_irqs);
    sensitive << extint_irq << eicr << eifr << eimsk << ext_int_inreg;

    SC_METHOD(output_data_mux);
    sensitive << mcucr << eimsk << eifr << eicr << mcusr << xdiv;
    sensitive << xdiv_sel << mcusr_sel << eicr_sel << eifr_sel << eimsk_sel << mcucr_sel;
  }
};

#endif
