//
// $Id: timer_counter.cc,v 1.11 2003/11/04 10:32:11 nj Exp $
//
#include "timer_counter.h"

//
// Synchronize 
//
void timer_counter::syncdffs()
{
  if (!reset)
    {
      ext1sa = SC_LOGIC_0;
      ext1sb = SC_LOGIC_0;
      ext2sa = SC_LOGIC_0;
      ext2sb = SC_LOGIC_0;
      tosc1sa = SC_LOGIC_0;
      tosc1sb = SC_LOGIC_0;
    }
  else
    {
      ext1sa = ext1;
      ext1sb = ext1sa.read();
      ext2sa = ext2;
      ext2sb = ext2sa.read();
      tosc1sa = tosc1;
      tosc1sb = tosc1sa.read();
    }
}

//
// Increment prescaler 1
//
void timer_counter::prescaler_1()
{
  sc_lv<10>   v_pre1cnt = pre1cnt;
  sc_uint<10> i_pre1cnt = v_pre1cnt;

  if (!reset)
    {
      v_pre1cnt = "0000000000";
      ck8 = SC_LOGIC_0;
      ck64 = SC_LOGIC_0;
      ck256 = SC_LOGIC_0;
      ck1024 = SC_LOGIC_0;
      ext1re = SC_LOGIC_0;
      ext1fe = SC_LOGIC_0;
      ext2re = SC_LOGIC_0;
      ext2fe = SC_LOGIC_0;
      ext1latched = SC_LOGIC_0;
      ext2latched = SC_LOGIC_0;
    }
  else
    {
      // Moved further down!
      // v_pre1cnt = v_pre1cnt.read() + 1;
      ck8 = ( ~ck8.read()) & ((v_pre1cnt[0] & v_pre1cnt[1]) & v_pre1cnt[2]);
      ck64 = ( ~ck64.read()) & (((((v_pre1cnt[0] & v_pre1cnt[1]) & v_pre1cnt[2]) & v_pre1cnt[3]) & v_pre1cnt[4]) & v_pre1cnt[5]);
      ck256 = ( ~ck256.read()) & (((((((v_pre1cnt[0] & v_pre1cnt[1]) & v_pre1cnt[2]) & v_pre1cnt[3]) & v_pre1cnt[4]) & v_pre1cnt[5]) & v_pre1cnt[6]) & v_pre1cnt[7]);
      ck1024 = ( ~ck1024.read()) & (((((((((v_pre1cnt[0] & v_pre1cnt[1]) & v_pre1cnt[2]) & v_pre1cnt[3]) & v_pre1cnt[4]) & v_pre1cnt[5]) & v_pre1cnt[6]) & v_pre1cnt[7]) & v_pre1cnt[8]) & v_pre1cnt[9]);
      ext1re = ( ~ext1re.read()) & (ext1sb.read() & ( ~ext1latched.read()));
      ext1fe = ( ~ext1fe.read()) & (( ~ext1sb.read()) & ext1latched.read());
      ext2re = ( ~ext2re.read()) & (ext2sb.read() & ( ~ext2latched.read()));
      ext2fe = ( ~ext2fe.read()) & (( ~ext2sb.read()) & ext2latched.read());
      ext1latched = ext1sb.read();
      ext2latched = ext2sb.read();

      // Here!
      v_pre1cnt = v_pre1cnt = i_pre1cnt + 1;
    }

  pre1cnt = v_pre1cnt;
}

void timer_counter::tosc1_rising_edge_detector()
{
  if (!reset)
    {
      tosc1re = SC_LOGIC_0;
      tosc1latched = SC_LOGIC_0;
    }
  else
    {
      tosc1re = ( ~tosc1re.read()) & (( ~tosc1latched.read()) & tosc1sb.read());
      tosc1latched = tosc1sb.read();
    }
}

//
// Increment prescaler 0
//
void timer_counter::prescaler_0_cnt()
{
  sc_lv<10> v_pre0cnt = pre0cnt;
  sc_uint<10> i_pre0cnt = v_pre0cnt;

  if (!reset)
    {
      v_pre0cnt = "0000000000";
    }
  else
    {
      if (pck0.read() == SC_LOGIC_1)
	{
	  v_pre0cnt = i_pre0cnt + 1;
#ifdef PRINTTC
  cout << "timer_counter::prescaler_0_cnt()" << endl;
  cout << "pre0cnt = " << v_pre0cnt << endl;
#endif
	}
    }
  pre0cnt = v_pre0cnt;
}

//
// Set prescaler 0 bit values
//
void timer_counter::prescaler_0()
{
  sc_lv<10> v_pre0cnt = pre0cnt.read();

  if (!reset)
    {
      pck08 = SC_LOGIC_0;
      pck032 = SC_LOGIC_0;
      pck064 = SC_LOGIC_0;
      pck0128 = SC_LOGIC_0;
      pck0256 = SC_LOGIC_0;
      pck01024 = SC_LOGIC_0;
      pck0_del = SC_LOGIC_0;
    }
  else
    {
      pck08 = (( ~pck08.read()) & (v_pre0cnt[0] & v_pre0cnt[1] & v_pre0cnt[2])) & pck0_del.read();
      pck032 = (( ~pck032.read()) & (v_pre0cnt[0] & v_pre0cnt[1] & v_pre0cnt[2] & v_pre0cnt[3] & v_pre0cnt[4])) & pck0_del.read();
      pck064 = (( ~pck064.read()) & v_pre0cnt[0] & v_pre0cnt[1] & v_pre0cnt[2] & v_pre0cnt[3] & v_pre0cnt[4] & v_pre0cnt[5]) & pck0_del.read();
      pck0128 = ( ~pck0128.read()) & ((v_pre0cnt[0] & v_pre0cnt[1] & v_pre0cnt[2] & v_pre0cnt[3] & v_pre0cnt[4] & v_pre0cnt[5] & v_pre0cnt[6]) & pck0_del.read());
      pck0256 = (( ~pck0256.read()) & (v_pre0cnt[0] & v_pre0cnt[1] & v_pre0cnt[2] & v_pre0cnt[3] & v_pre0cnt[4] & v_pre0cnt[5] & v_pre0cnt[6] & v_pre0cnt[7])) & pck0_del.read();
      pck01024 = (( ~pck01024.read()) & (v_pre0cnt[0] & v_pre0cnt[1] & v_pre0cnt[2] & v_pre0cnt[3] & v_pre0cnt[4] & v_pre0cnt[5] & v_pre0cnt[6] & v_pre0cnt[7] & v_pre0cnt[8] & v_pre0cnt[9])) & pck0_del.read();
      pck0_del = pck0.read();
    }
}

//
// attention !!! this model only emulates the asynchronous mode of tcnt0. 
// real tcnt0 of atmega103/603 operates with two separate clock sources.
//
void timer_counter::counter_0()
{
  if (!reset)
    {
      tcnt0 = "00000000";
    }
  else
    {
      if (((tcnt0_en.read() | tcnt0_clr.read()) | tcnt0_ld_imm.read()) == SC_LOGIC_1)
      {
        tcnt0 = tcnt0_in.read();
      }
    }
}

void timer_counter::counter_0_inc()
{
  if (!reset)
    {
      tcnt0_inc = SC_LOGIC_0;
    }
  else
    {
      tcnt0_inc = (( ~tcnt0_inc.read()) & (tcnt0_en.read() & ( ~tcnt0_ld_imm.read()))) | (tcnt0_inc.read() & ( ~tcnt0_ld_imm.read()));
    }
}

void timer_counter::tccr0_control()
{
  sc_lv<8> v_tcnt0 = tcnt0;
  sc_lv<8> v_tccr0 = tccr0;
  sc_lv<8> v_assr = assr;
  sc_lv<8> v_dbus_in = dbus_in;

  if (!reset)
    {
      tcnt0_cnt_dir = SC_LOGIC_0;
      tcr0ub_tmp = SC_LOGIC_0;
      ocr0ub_tmp = SC_LOGIC_0;
      tcn0ub_tmp = SC_LOGIC_0;
      assr = "00000000";
      // assr = "00001000";
    }
  else
    {
      tcnt0_cnt_dir = (( ~tcnt0_cnt_dir.read()) & (((((((((v_tcnt0[7] & v_tcnt0[6]) & v_tcnt0[5]) & v_tcnt0[4]) & v_tcnt0[3]) & v_tcnt0[2]) & v_tcnt0[1]) & ( ~v_tcnt0[0])) & tcnt0_en.read()) & pwm0)) | (tcnt0_cnt_dir.read() & ( ~(((((((((( ~v_tcnt0[7]) & ( ~v_tcnt0[6])) & ( ~v_tcnt0[5])) & ( ~v_tcnt0[4])) & ( ~v_tcnt0[3])) & ( ~v_tcnt0[2])) & ( ~v_tcnt0[1])) & v_tcnt0[0]) & tcnt0_en.read()) | ( ~pwm0))));
      tcr0ub_tmp = (( ~tcr0ub_tmp.read()) & ((tccr0_sel.read() & iowe) & as0)) | (tcr0ub_tmp.read() & ( ~pck0.read()));
      ocr0ub_tmp = (( ~ocr0ub_tmp.read()) & ((ocr0_sel.read() & iowe) & as0)) | (ocr0ub_tmp.read() & ( ~pck0.read()));
      tcn0ub_tmp = (( ~tcn0ub_tmp.read()) & ((tcnt0_sel.read() & iowe) & as0)) | (tcn0ub_tmp.read() & ( ~pck0.read()));
      tcr0ub = (( ~tcr0ub) & ((tccr0_sel.read() & iowe) & as0)) | (tcr0ub & ( ~(( ~tcr0ub_tmp.read()) & pck0.read())));
      ocr0ub = (( ~ocr0ub) & ((ocr0_sel.read() & iowe) & as0)) | (ocr0ub & ( ~(( ~ocr0ub_tmp.read()) & pck0.read())));
      tcn0ub = (( ~tcn0ub) & ((tcnt0_sel.read() & iowe) & as0)) | (tcn0ub & ( ~(( ~tcn0ub_tmp.read()) & pck0.read())));
      as0 = (( ~as0) & ((assr_sel.read() & iowe) & v_dbus_in[3])) | (as0 & ( ~((assr_sel.read() & iowe) & ( ~v_dbus_in[3]))));
    }
}

void timer_counter::tcnt0_temp_control()
{
  sc_lv<8> v_assr = assr;

  if (!reset)
    {
      tccr0_tmp = "00000000";
    }
  else
    {
      if ((tccr0_sel.read() & iowe & as0) == SC_LOGIC_1)
	{
	  tccr0_tmp = dbus_in;
#ifdef PRINTTC
  cout << "timer_counter::tcnt0_temp_control()" << endl;
  cout << "tccr0_tmp = " << dbus_in.read() << endl;
#endif
	}
    }
}

void timer_counter::tcnt0_temp_cnt()
{
  sc_lv<8> v_assr = assr;

  if (!reset)
    {
      tcnt0_tmp = "00000000";
    }
  else
    {
      if (((tcnt0_sel.read() & iowe) & as0) == SC_LOGIC_1)
	{
	  tcnt0_tmp = dbus_in;
#ifdef PRINTTC
  cout << "timer_counter::tcnt0_temp_cnt()" << endl;
  cout << "tcnt0_tmp = " << dbus_in.read() << endl;
#endif
	}
    }
}

void timer_counter::tcnt0_temp_compare()
{
  if (!reset)
    {
      ocr0_tmp = "00000000";
    }
  else
    {
      if ((ocr0_sel.read() & iowe) == SC_LOGIC_1)
	{
	  ocr0_tmp = dbus_in;
#ifdef PRINTTC
  cout << "timer_counter::tcnt0_temp_compare()" << endl;
  cout << "ocr0_tmp = " << dbus_in.read() << endl;
#endif
	}
    }
}

void timer_counter::tcnt0_control()
{
  sc_lv<8> v_tccr0 = tccr0;
  sc_lv<8> v_tccr0_in = tccr0_in;
  sc_lv<8> v_assr = assr;

  if (!reset)
    {
      v_tccr0 = "00000000";
    }
  else
    {
      if ((((tccr0_sel.read() & iowe) & ~as0) | (((tcr0ub & ~tcr0ub_tmp.read()) & pck0.read()) & as0)) == SC_LOGIC_1)
	{
	  v_tccr0[6] = v_tccr0_in[6];
	  v_tccr0[5] = v_tccr0_in[5];
	  v_tccr0[4] = v_tccr0_in[4];
	  v_tccr0[3] = v_tccr0_in[3];
	  v_tccr0[2] = v_tccr0_in[2];
	  v_tccr0[1] = v_tccr0_in[1];
	  v_tccr0[0] = v_tccr0_in[0];
#ifdef PRINTTC
  cout << "timer_counter::tcnt0_control()" << endl;
  cout << "tccr0 = " << v_tccr0 << endl;
#endif
	}
    }
  tccr0 = v_tccr0;
}

void timer_counter::tcnt0_compare()
{
  sc_lv<8> v_tccr0 = tccr0;
  sc_lv<8> v_tcnt0 = tcnt0;
  sc_lv<8> v_assr = assr;

  if (!reset)
    {
      ocr0 = "00000000";
    }
  else
    {
      if ((((((ocr0_sel.read() & iowe) & ( ~as0)) | (((ocr0ub & ( ~ocr0ub_tmp.read())) & pck0.read()) & as0)) & ( ~pwm0)) | (((((((((v_tcnt0[7] & v_tcnt0[6]) & v_tcnt0[5]) & v_tcnt0[4]) & v_tcnt0[3]) & v_tcnt0[2]) & v_tcnt0[1]) & v_tcnt0[0]) & pwm0) & tcnt0_en.read())) == SC_LOGIC_1)
	{
	  ocr0 = ocr0_in.read();
#ifdef PRINTTC
	  cout << "timer_counter::tcnt0_compare()" << endl;
	  cout << "ocr0 = " << ocr0_in.read() << endl;
#endif
	}
    }
}

void timer_counter::output_tcnt0()
{
  sc_lv<8> v_tccr0 = tccr0;

  if (!reset)
    {
      oc0_pwm0_int = SC_LOGIC_0;
    }
  else
    {
      // sc_logic __case_expression__ = oc0_pwm0_int.read();
      if (oc0_pwm0_int.read() == SC_LOGIC_0)
	{
	  if (((
		(( ~pwm0) & com00) & tcnt0_cmp_out.read()) == SC_LOGIC_1) | 
	      (
	       (((pwm0 & com01) & tcnt0_en.read()) == SC_LOGIC_1) & 
	       (((((((tcnt0.read() == "11111111") & (com00 == SC_LOGIC_1)) & (ocr0_tmp.read() == "00000000")) | 
		   (((tcnt0.read() == "11111111") & (com00 == SC_LOGIC_0)) & (ocr0_tmp.read() == "11111111")))|
		  (((tcnt0.read() == "11111111") & (com00 == SC_LOGIC_1)) & (ocr0_tmp.read() != "11111111"))) | 
		 (((((tcnt0.read() != "11111111") & (tcnt0_cnt_dir.read() == SC_LOGIC_1)) & (com00 == SC_LOGIC_0)) & (ocr0_tmp.read() != "00000000")) & (ocr0.read() == tcnt0.read()))) | 
		(((((tcnt0.read() != "11111111") & (tcnt0_cnt_dir.read() == SC_LOGIC_0)) & (com00 == SC_LOGIC_1)) & (ocr0_tmp.read() != "00000000")) & (ocr0.read() == tcnt0.read())))))
	    {
	      oc0_pwm0_int = SC_LOGIC_1;
	    }
	}
      else if (oc0_pwm0_int.read() == SC_LOGIC_1)
	{
	  if ((((( ~pwm0) & (com01 | com00)) & tcnt0_cmp_out.read()) == SC_LOGIC_1) | 
	      ((((pwm0 & com01) & tcnt0_en.read()) == SC_LOGIC_1) & 
	       (((((((tcnt0.read() == "11111111") & (com00 == SC_LOGIC_1)) & (ocr0_tmp.read() == "11111111")) | 
		   (((tcnt0.read() == "11111111") & (com00 == SC_LOGIC_0)) & (ocr0_tmp.read() == "00000000"))) | 
		  (((tcnt0.read() == "11111111") & (com00 == SC_LOGIC_0)) & (ocr0_tmp.read() != "11111111"))) | 
		 (((((tcnt0.read() != "11111111") & (tcnt0_cnt_dir.read() == SC_LOGIC_0)) & (com00 == SC_LOGIC_0)) & (ocr0_tmp.read() != "00000000")) & (ocr0.read() == tcnt0.read()))) | 
		(((((tcnt0.read() != "11111111") & (tcnt0_cnt_dir.read() == SC_LOGIC_1)) & (com00 == SC_LOGIC_1)) & (ocr0_tmp.read() != "00000000")) & (ocr0.read() == tcnt0.read())))))
	    {
	      oc0_pwm0_int = SC_LOGIC_0;
	    }
	    // else {}
	}
    }
}

void timer_counter::counter_2()
{
  if (!reset)
    {
      tcnt2 = "00000000";
    }
  else
    {
      if (((tcnt2_en.read() | tcnt2_clr.read()) | tcnt2_ld_imm.read()) == SC_LOGIC_1)
	{
	  tcnt2 = tcnt2_in.read();
	}
    }
}

void timer_counter::counter_2_inc()
{
  if (!reset)
    {
      tcnt2_inc = SC_LOGIC_0;
    }
  else
    {
      tcnt2_inc = (( ~tcnt2_inc.read()) & (tcnt2_en.read() & ( ~(tcnt2_sel.read() & iowe)))) | (tcnt2_inc.read() & ( ~(tcnt2_sel.read() & iowe)));
    }
}

void timer_counter::tccr2_control()
{
  sc_lv<8> v_tcnt2 = tcnt2;
  sc_lv<8> v_tccr2 = tccr2;

  if (!reset)
    {
      tcnt2_cnt_dir = SC_LOGIC_0;
    }
  else
    {
      tcnt2_cnt_dir = (( ~tcnt2_cnt_dir.read()) & (((((((((v_tcnt2[7] & v_tcnt2[6]) & v_tcnt2[5]) & v_tcnt2[4]) & v_tcnt2[3]) & v_tcnt2[2]) & v_tcnt2[1]) & ( ~v_tcnt2[0])) & tcnt2_en.read()) & pwm2)) | (tcnt2_cnt_dir.read() & ( ~(((((((((( ~v_tcnt2[7]) & ( ~v_tcnt2[6])) & ( ~v_tcnt2[5])) & ( ~v_tcnt2[4])) & ( ~v_tcnt2[3])) & ( ~v_tcnt2[2])) & ( ~v_tcnt2[1])) & v_tcnt2[0]) & tcnt2_en.read()) | ( ~pwm2))));
    }
}

void timer_counter::tcnt2_temp_compare()
{
  if (!reset)
    {
      ocr2_tmp = "00000000";
    }
  else
    {
      if ((ocr2_sel.read() & iowe) == SC_LOGIC_1)
	{
	  ocr2_tmp = dbus_in;
	}
    }
}

void timer_counter::tcnt2_control()
{
  sc_lv<8> v_tccr2 = tccr2;
  sc_lv<8> v_dbus_in = dbus_in;

  if (!reset)
    {
      v_tccr2 = "00000000";
    }
  else
    {
      if ((tccr2_sel.read() & iowe) == SC_LOGIC_1)
	{
	  v_tccr2[6] = v_dbus_in[6];
	  v_tccr2[5] = v_dbus_in[5];
	  v_tccr2[4] = v_dbus_in[4];
	  v_tccr2[3] = v_dbus_in[3];
	  v_tccr2[2] = v_dbus_in[2];
	  v_tccr2[1] = v_dbus_in[1];
	  v_tccr2[0] = v_dbus_in[0];
	}
    }
  tccr2 = v_tccr2;
}

//
// Write Output Compare Register 2 from input
//
void timer_counter::tcnt2_compare()
{
  sc_lv<8> v_tcnt2 = tcnt2;
  sc_lv<8> v_tccr2 = tccr2;

  if (!reset)
    {
      ocr2 = "00000000";
    }
  else
    {
      if (((ocr2_sel.read() & iowe) | (((((((((v_tcnt2[7] & v_tcnt2[6]) & v_tcnt2[5]) & v_tcnt2[4]) & v_tcnt2[3]) & v_tcnt2[2]) & v_tcnt2[1]) & v_tcnt2[0]) & pwm2) & tcnt2_en.read())) == SC_LOGIC_1)
	{
	  ocr2 = ocr2_in.read();
	}
    }
}

//
// Write Output Compare register 2 interrupt
//
void timer_counter::output_tcnt2()
{
  sc_lv<8> v_tccr2 = tccr2;

  if (!reset)
    {
      oc2_pwm2_int = SC_LOGIC_0;
    }
  else
    {
      if (oc2_pwm2_int.read() == SC_LOGIC_0)
	{
	  if ((((( ~pwm2) & com20) & tcnt2_cmp_out.read()) == SC_LOGIC_1) | ((((pwm2 & com21) & tcnt2_en.read()) == SC_LOGIC_1) & 
(((((((tcnt2.read() == "11111111") & (com20 == SC_LOGIC_1)) & (ocr2_tmp.read() == "00000000")) | 
    (((tcnt2.read() == "11111111") & (com20 == SC_LOGIC_0)) & (ocr2_tmp.read() == "11111111"))) | 
    (((tcnt2.read() == "11111111") & (com20 == SC_LOGIC_1)) & (ocr2_tmp.read() != "11111111"))) | 
  (((((tcnt2.read() != "11111111") & (tcnt2_cnt_dir.read() == SC_LOGIC_1)) & (com20 == SC_LOGIC_0)) & (ocr2_tmp.read() != "00000000")) & (ocr2.read() == tcnt2.read()))) | 
  (((((tcnt2.read() != "11111111") & (tcnt2_cnt_dir.read() == SC_LOGIC_0)) & (com20 == SC_LOGIC_1)) & (ocr2_tmp.read() != "00000000")) & (ocr2.read() == tcnt2.read())))))
	    {
	      oc2_pwm2_int = SC_LOGIC_1;
	    }
	}
      else if (oc2_pwm2_int.read() == SC_LOGIC_1)
	{
	  if ((((( ~pwm2) & (com21 | com20)) & tcnt2_cmp_out.read()) == SC_LOGIC_1) | ((((pwm2 & com21) & tcnt2_en.read()) == SC_LOGIC_1) & 
(((((((tcnt2.read() == "11111111") & (com20 == SC_LOGIC_1)) & (ocr2_tmp.read() == "11111111")) | 
    (((tcnt2.read() == "11111111") & (com20 == SC_LOGIC_0)) & (ocr2_tmp.read() == "00000000"))) | 
    (((tcnt2.read() == "11111111") & (com20 == SC_LOGIC_0)) & (ocr2_tmp.read() != "11111111"))) | 
  (((((tcnt2.read() != "11111111") & (tcnt2_cnt_dir.read() == SC_LOGIC_0)) & (com20 == SC_LOGIC_0)) & (ocr2_tmp.read() != "00000000")) & (ocr2.read() == tcnt0.read()))) | 
  (((((tcnt2.read() != "11111111") & (tcnt2_cnt_dir.read() == SC_LOGIC_1)) & (com20 == SC_LOGIC_1)) & (ocr2_tmp.read() != "00000000")) & (ocr2.read() == tcnt0.read())))))
	    {
	      oc2_pwm2_int = SC_LOGIC_0;
	    }
	}
      // else {}
    }
}

//
// Set bits in the Timer Interrupt Flag Register
//
void timer_counter::tifr_bits()
{
  sc_lv<8> v_tcnt0 = tcnt0.read();
  sc_lv<8> v_tccr0 = tccr0.read();
  sc_lv<8> v_tccr2 = tccr2.read();
  sc_lv<8> v_dbus_in = dbus_in.read();
  sc_lv<8> v_tcnt2 = tcnt2.read();

   // For outputs - bits are #define's
  sc_lv<8> v_tifr = tifr.read();

  if (!reset)
    {
      v_tifr = "00000000";
    }
  else
    {
      tov0 = (( ~tov0) & (tcnt0_en.read() & ((~pwm0 & v_tcnt0[7] & v_tcnt0[6] & v_tcnt0[5] & v_tcnt0[4] & v_tcnt0[3] & v_tcnt0[2] & v_tcnt0[1] & v_tcnt0[0]) | (((pwm0 & ( ~((((((v_tcnt0[7] | v_tcnt0[6]) | v_tcnt0[5]) | v_tcnt0[4]) | v_tcnt0[3]) | v_tcnt0[2]) | v_tcnt0[1]))) & v_tcnt0[0]) & tcnt0_cnt_dir.read())))) | (tov0 & ( ~(tc0ovfirq_ack | (tifr_sel.read() & iowe.read() & v_dbus_in[0]))));
      // ocf0 = (( ~ocf0) & tcnt0_en.read() & tcnt0_cmp_out.read()) | (ocf0 & ( ~(tc0cmpirq_ack | (tifr_sel.read() & iowe & v_dbus_in[1]))));
      ocf0 = (( ~ocf0) & tcnt0_en2.read() & tcnt0_cmp_out.read()) | (ocf0 & ( ~(tc0cmpirq_ack | (tifr_sel.read() & iowe & v_dbus_in[1]))));
      tov2 = (( ~tov2) & (tcnt2_en.read() & (((((((((( ~pwm2) & v_tcnt2[7]) & v_tcnt2[6]) & v_tcnt2[5]) & v_tcnt2[4]) & v_tcnt2[3]) & v_tcnt2[2]) & v_tcnt2[1]) & v_tcnt2[0]) | (((pwm2 & ( ~((((((v_tcnt2[7] | v_tcnt2[6]) | v_tcnt2[5]) | v_tcnt2[4]) | v_tcnt2[3]) | v_tcnt2[2]) | v_tcnt2[1]))) & v_tcnt2[0]) & tcnt2_cnt_dir.read())))) | (tov2 & ( ~(tc2ovfirq_ack | ((tifr_sel.read() & iowe) & v_dbus_in[6]))));
      ocf2 = ((( ~ocf2) & tcnt2_en.read()) & tcnt2_cmp_out.read()) | (ocf2 & ( ~(tc2cmpirq_ack | ((tifr_sel.read() & iowe) & v_dbus_in[7]))));
    }

  tifr = v_tifr;
#ifdef PRINTTC
  /*  if (tifr_sel.read() == SC_LOGIC_1)
    {
  */
      cout << "timer_counter::tifr_bits()" << endl;
      cout << "tifr_sel = " << tifr_sel.read() << endl;
      cout << "tcnt0 = " << tcnt0.read() << endl;
      cout << "tccr0 = " << tccr0.read() << endl;
      cout << "tccr2 = " << tccr2.read() << endl;
      cout << "dbus_in = " << dbus_in.read() << endl;
      cout << "tcnt2 = " << tcnt2.read() << endl;
      cout << "tc0ovfirq_ack = " << tc0ovfirq_ack << endl;
      cout << "tcnt0_en = " << tcnt0_en.read() << endl;
      cout << "iowe = " << iowe.read() << endl;
      cout << "tifr = " << v_tifr << endl;
      //  }
#endif
}

//
// Read values in to the Timer Interrupt MaSK register
//
void timer_counter::timsk_bits()
{
  sc_lv<8> v_timsk = timsk.read();

  if (!reset)
    {
      // v_timsk = "00000000";
      v_timsk = "00000010";
    }
  else
    {
      if ((timsk_sel.read() & iowe) == SC_LOGIC_1)
	{
	  v_timsk = dbus_in;
#ifdef PRINTTC
  cout << "timer_counter::timsk_bits()" << endl;
  cout << "timsk = " << v_timsk << endl;
#endif
	}
    }
  timsk = v_timsk;
}

//
// Concurrent statements
//
// sensitive << assr << tccr0 << tccr1a << tccr2 << tcnt0 << tcnt2 << tccr0_tmp
// sensitive << ocr0 << ocr0_tmp << ocr2 << tifr << timsk
// sensitive << oc0_pwm0_int << oc2_pwm2_int << iore << iowe << adr
// sensitive << tccr0_sel << tccr1a_sel << tccr1b_sel << tccr2_sel << assr_sel
// sensitive << timsk_sel << tifr_sel << tcnt0_sel << tcnt2_sel << ocr0_sel << ocr2_sel
// sensitive << tcnt1h_sel << tcnt1l_sel << ocr1ah_sel << ocr1al_sel << ocr1bh_sel << ocr1bl_sel 
// sensitive << icr1ah_sel << icr1al_sel
// sensitive << tosc1re << pck0 << tcn0ub_tmp << tcnt0_ld_imm << tcnt0_inc << tcnt0_cmp_out
// sensitive << tcnt0_clr << tcnt0_cnt_dir
// sensitive << dbus_in
// sensitive << pck08 << pck032 << pck064 << pck0128 << pck0256 << pck01024
void timer_counter::concurrent_statements()
{
  // sc_lv<8> v_cnt0 = cnt0;
  // sc_lv<8> v_cnt2 = cnt2;
  sc_lv<6> v_adr = adr.read();

  sc_lv<8> v_assr = assr;

  sc_lv<8> v_tccr0 = tccr0;
  sc_lv<8> v_tccr1a = tccr1a;
  sc_lv<8> v_tccr2 = tccr2;

  sc_lv<8> v_tcnt0 = tcnt0;
  sc_uint<8> i_tcnt0 = v_tcnt0;
  sc_lv<8> v_tcnt2 = tcnt2;
  sc_uint<8> i_tcnt2 = v_tcnt2;

  sc_lv<8> v_tccr0_tmp = tccr0_tmp;
  sc_lv<8> v_ocr0 = ocr0;
  sc_lv<8> v_ocr0_tmp = ocr0_tmp;
  sc_lv<8> v_ocr2 = ocr2;
  sc_lv<8> v_tifr = tifr;
  sc_lv<8> v_timsk = timsk;

  // Output
  sc_lv<8> v_tcnt0_in;
  sc_lv<8> v_tcnt2_in;
  sc_lv<8> v_dbus_out;

  tcnt1_en = (~cs12 & ~cs11 & cs10) | // ck            "001"
    (ck8 & ~cs12 & cs11 & ~cs10) |    // ck/8          "010"
    (ck64 & ~cs12 & cs11 & cs10) |    // ck/64         "011"
    (ck256 & cs12 & ~cs11 & ~cs10) |  // ck/256        "100"
    (ck1024 & cs12 & ~cs11 & cs10) |  // ck/1024       "101"
    (ext1fe & cs12 & cs11 & ~cs10) |  // falling edge  "110"
    (ext1re & cs12 & cs11 & cs10);    // rising edge   "111"
			
  tcnt2_en = (~cs22 & ~cs21 & cs20) | // ck            "001"
    (ck8 & ~cs22 & cs21 & ~cs20) |    // ck/8          "010"
    (ck64 & ~cs22 & cs21 & cs20) |    // ck/64         "011"
    (ck256 & cs22 & ~cs21 & ~cs20) |  // ck/256        "100"
    (ck1024 & cs22 & ~cs21 & cs20) |  // ck/1024       "101"
    (ext2fe & cs22 & cs21 & ~cs20) |  // falling edge  "110"
    (ext2re & cs22 & cs21 & cs20);    // rising edge   "111"

  if (as0 == SC_LOGIC_1)
    pck0 = tosc1re;
  else
    pck0 = SC_LOGIC_1;

  // Problem here!!
  tcnt0_en = (pck0 & ~cs02 & ~cs01 & cs00) |  // pck       "001" !!??
    (pck08 & ~cs02 & cs01 & ~cs00) |          // pck/8     "010"
    (pck032 & ~cs02 & cs01 & cs00) |          // pck/32    "011"
    (pck064 & cs02 & ~cs01 & ~cs00) |         // pck/64    "100"
    (pck0128 & cs02 & ~cs01 & cs00) |         // pck/64    "101"
    (pck0256 & cs02 & cs01 & ~cs00) |         // pck/256   "110"
    (pck01024 & cs02 & cs01 & cs00);          // pck/1024  "111"

  // i/o address decoder
  tccr0_sel = SC_LOGIC_0;
  tccr1a_sel = SC_LOGIC_0;
  tccr1b_sel = SC_LOGIC_0;
  tccr2_sel  = SC_LOGIC_0;
  assr_sel   = SC_LOGIC_0;
  timsk_sel  = SC_LOGIC_0;
  tifr_sel   = SC_LOGIC_0;
  tcnt0_sel  = SC_LOGIC_0;
  tcnt2_sel  = SC_LOGIC_0;
  ocr0_sel   = SC_LOGIC_0;
  ocr2_sel   = SC_LOGIC_0;
  tcnt1h_sel = SC_LOGIC_0;
  tcnt1l_sel = SC_LOGIC_0;
  ocr1ah_sel = SC_LOGIC_0;
  ocr1al_sel = SC_LOGIC_0;
  ocr1bh_sel = SC_LOGIC_0;
  ocr1bl_sel = SC_LOGIC_0;
  icr1ah_sel = SC_LOGIC_0;
  icr1al_sel = SC_LOGIC_0;

  if (v_adr==TCCR0_Address)  tccr0_sel  = SC_LOGIC_1;
  if (v_adr==TCCR1A_Address) tccr1a_sel = SC_LOGIC_1;
  if (v_adr==TCCR1B_Address) tccr1b_sel = SC_LOGIC_1;
  if (v_adr==TCCR2_Address)  tccr2_sel  = SC_LOGIC_1;
  if (v_adr==ASSR_Address)   assr_sel   = SC_LOGIC_1;
  if (v_adr==TIMSK_Address)  timsk_sel  = SC_LOGIC_1;
  if (v_adr==TIFR_Address)   tifr_sel   = SC_LOGIC_1;
  if (v_adr==TCNT0_Address)  tcnt0_sel  = SC_LOGIC_1;
  if (v_adr==TCNT2_Address)  tcnt2_sel  = SC_LOGIC_1;
  if (v_adr==OCR0_Address)   ocr0_sel   = SC_LOGIC_1;
  if (v_adr==OCR2_Address)   ocr2_sel   = SC_LOGIC_1;
  if (v_adr==TCNT1H_Address) tcnt1h_sel = SC_LOGIC_1;
  if (v_adr==TCNT1L_Address) tcnt1l_sel = SC_LOGIC_1;
  if (v_adr==OCR1AH_Address) ocr1ah_sel = SC_LOGIC_1;
  if (v_adr==OCR1AL_Address) ocr1al_sel = SC_LOGIC_1;
  if (v_adr==OCR1BH_Address) ocr1bh_sel = SC_LOGIC_1;
  if (v_adr==OCR1BL_Address) ocr1bl_sel = SC_LOGIC_1;
  if (v_adr==ICR1AH_Address) icr1ah_sel = SC_LOGIC_1;
  if (v_adr==ICR1AL_Address) icr1al_sel = SC_LOGIC_1;

  if (tcnt0_ld_imm == SC_LOGIC_1)
    v_tcnt0_in = tcnt0_imm_in;          // immediate value (from dbus_in or tcnt0_tmp)
  else if (tcnt0_clr == SC_LOGIC_1)
    v_tcnt0_in = "00000000";            // synchronous clear (for ocr)
  else if (tcnt0_cnt_dir == SC_LOGIC_1)
    v_tcnt0_in = v_tcnt0 = i_tcnt0 - 1; // decrement (for pwm)
  else
    v_tcnt0_in = v_tcnt0 = i_tcnt0 + 1; // increment (for counter and pwm)

  if ((tcn0ub & ~tcn0ub_tmp.read() & pck0 & as0) == SC_LOGIC_1)
    tcnt0_imm_in = tcnt0_tmp;
  else
    tcnt0_imm_in = dbus_in;

  tcnt0_clr = tcnt0_cmp_out & ctc0 & ~pwm0;

  if ((tcnt0.read()==ocr0.read()) && !(ocr0.read()=="00000000") && (tcnt0_inc.read()==SC_LOGIC_1))
    tcnt0_cmp_out = SC_LOGIC_1;
  else
    tcnt0_cmp_out = SC_LOGIC_0;

  tcnt0_ld_imm = (tcnt0_sel & iowe & ~as0) | (tcn0ub & ~tcn0ub_tmp.read() & pck0 & as0);

  if ((tcr0ub & ~tcr0ub_tmp.read() & pck0 & as0) == SC_LOGIC_1)
    {
      tccr0_in = tccr0_tmp;
      /*
      cout << "timer_counter::concurrent_statements()" << endl;
      cout << "tccr0_in = tccr0_tmp (" << tccr0_tmp.read() << ")" << endl;
      */
    }
  else
    {
      tccr0_in = dbus_in;
      /*
      cout << "timer_counter::concurrent_statements()" << endl;
      cout << "tccr0_in = dbus_in (" << dbus_in.read() << ")" << endl;
      */
    }

  // ocr0 can be loaded from ocr0_tmp (asynchronous non pwm mode, synchronous pwm mode )
  // or from dbus_in	(synchronous non pwm mode only)
  if (((tcr0ub & ~tcr0ub_tmp.read() & pck0 & as0 & ~pwm0) | (v_tcnt0[7] & v_tcnt0[6] & v_tcnt0[5] & v_tcnt0[4] & v_tcnt0[3] & v_tcnt0[2] & v_tcnt0[1] & v_tcnt0[0] & pwm0 & tcnt0_en)) == SC_LOGIC_1) 
    ocr0_in = ocr0_tmp;
  else
    ocr0_in = dbus_in;

  oc0_pwm0 = oc0_pwm0_int;

  tcnt2_ld_imm = tcnt2_sel & iowe; // write to tcnt2

  // input multiplexer of tccr2 register
  if (tcnt2_ld_imm == SC_LOGIC_1)
    v_tcnt2_in = dbus_in; // immediate value (from dbus_in)
  else if (tcnt2_clr == SC_LOGIC_1)
    v_tcnt2_in = "00000000"; // synchronous clear (for ocr)
  else if (tcnt2_cnt_dir == SC_LOGIC_1)
    v_tcnt2_in = i_tcnt2-1; //  decrement (for pwm)
  else 
    v_tcnt2_in = i_tcnt2+1; // icrement (for counter and pwm)

  tcnt2_clr = tcnt2_cmp_out & ctc2 & ~pwm2;

  if ((tcnt2.read()==ocr2.read()) && !(ocr2.read() == "00000000") && (tcnt2_inc.read() == SC_LOGIC_1))
    tcnt2_cmp_out = SC_LOGIC_1;
  else 
    tcnt2_cmp_out = SC_LOGIC_0;

  // ocr2 can be loaded from ocr2_tmp (pwm mode ) or from dbus_in (non pwm mode only)
  if (((v_tcnt2[7] & v_tcnt2[6] & v_tcnt2[5] & v_tcnt2[4] & v_tcnt2[3] & v_tcnt2[2] & v_tcnt2[1] & v_tcnt2[0] & pwm2 & tcnt2_en)) == SC_LOGIC_1)
    ocr2_in = ocr2_tmp;
  else
    ocr2_in = dbus_in;

  oc2_pwm2 = oc2_pwm2_int;

  //
  // COMMON INTERRUPT/CONTROL BITS
  //

  // interrupt flags of timer/counter0
  tc0ovfirq = tov0 & toie0;	  // interrupt on overflow of tcnt0
  // tc0ovfirq = v_tifr[0] & v_timsk[0]

  if ((tov0 & toie0) == SC_LOGIC_1)
    {
      cout << "tc0ovfirq = 1" << endl;
      // sc_stop();
    }
  tc0cmpirq = ocf0 & ocie0;	  // interrupt on compare match	of tcnt0
  // tc0cmpirq = v_tifr[1] & v_timsk[1]

  // interrupt flags of timer/counter2
  tc2ovfirq = tov2 & toie2;	  // interrupt on overflow of tcnt2
  tc2cmpirq = ocf2 & ocie2;	  // interrupt on compare match	of tcnt2

  //
  // BUS INTERFACES
  //
  out_en = (tccr0_sel | tccr1a_sel | tccr1b_sel | tccr2_sel | assr_sel | timsk_sel | 
	    tifr_sel  | tcnt0_sel  | tcnt2_sel | ocr0_sel | ocr2_sel | tcnt1h_sel | 
	    tcnt1l_sel | ocr1ah_sel | ocr1al_sel | ocr1bh_sel | ocr1bl_sel | icr1ah_sel |  
	    icr1al_sel) & iore;

  COMMON_MUX_GENERATE(0);
  COMMON_MUX_GENERATE(1);
  COMMON_MUX_GENERATE(2);
  COMMON_MUX_GENERATE(3);
  COMMON_MUX_GENERATE(4);
  COMMON_MUX_GENERATE(5);
  COMMON_MUX_GENERATE(6);
  COMMON_MUX_GENERATE(7);

  // Outputs
  tcnt0_in = v_tcnt0_in;
  tcnt2_in = v_tcnt2_in;
  dbus_out = v_dbus_out;

  // Static signal assignments
  oc0_pwm0   = SC_LOGIC_0;
  oc1a_pwm1a = SC_LOGIC_0;
  oc1b_pwm1b = SC_LOGIC_0;
  oc2_pwm2   = SC_LOGIC_0;

  tc1cmpairq = SC_LOGIC_0;
  tc1cmpbirq = SC_LOGIC_0;
  tc1icirq   = SC_LOGIC_0;
  tc1ovfirq  = SC_LOGIC_0;
}


//
// Synchronous method so tcnt0_en2 will be set one clock cycle after tcnt0_en
//
void timer_counter::set_tcnt0_en()
{
  sc_lv<8> v_tccr0 = tccr0;

  if (tcnt0_en.read() == SC_LOGIC_1)
    tcnt0_en2 = SC_LOGIC_1;
  else
    tcnt0_en2 = SC_LOGIC_0;

  /*
  tcnt0_en = (pck0 & ~cs02 & ~cs01 & cs00) |  // pck       "001" !!??
    (pck08 & ~cs02 & cs01 & ~cs00) |          // pck/8     "010"
    (pck032 & ~cs02 & cs01 & cs00) |          // pck/32    "011"
    (pck064 & cs02 & ~cs01 & ~cs00) |         // pck/64    "100"
    (pck0128 & cs02 & ~cs01 & cs00) |         // pck/64    "101"
    (pck0256 & cs02 & cs01 & ~cs00) |         // pck/256   "110"
    (pck01024 & cs02 & cs01 & cs00);          // pck/1024  "111"
  */
}
