//
// $Id: timer_counter.h,v 1.10 2003/11/05 16:16:59 nj Exp $
//
#ifndef _TIMER_COUNTER_H
#define _TIMER_COUNTER_H
#include <systemc.h>
#include "config.h"
#include "constants.h"

#if defined(PRINTTC) && defined(SYNTHESIS)
#undef PRINTTC
#else
#define PRINTTC 1
#endif

#include "avrucpackage.h"

#define COMMON_MUX_GENERATE(x) v_dbus_out[x] = (v_tccr0[x] & (tccr0_sel & ~as0)) | \
  (v_tccr0_tmp[x] & (tccr0_sel & as0)) | \
  (v_ocr0[x] & (ocr0_sel & ~as0)) | \
  (v_ocr0_tmp[x] & (ocr0_sel & as0)) | \
  (v_tcnt0[x] & tcnt0_sel) | \
  (v_tccr2[x] & tccr2_sel) | \
  (v_ocr2[x] & ocr2_sel) | \
  (v_tcnt2[x] & tcnt2_sel) | \
  (v_tifr[x] & tifr_sel) | \
  (v_timsk[x] & timsk_sel);

// tccr0 bits
// timer/counter control register for counter 0
#define cs00  v_tccr0[0]
#define cs01  v_tccr0[1]
#define cs02  v_tccr0[2]
#define ctc0  v_tccr0[3]
#define com00 v_tccr0[4]
#define com01 v_tccr0[5]
#define pwm0  v_tccr0[6]

// tccr1a bits
#define pwm10  v_tccr1a[0]
#define pwm11  v_tccr1a[1]
#define com1b0 v_tccr1a[4]
#define com1b1 v_tccr1a[5]
#define com1a0 v_tccr1a[4]
#define com1a1 v_tccr1a[5]

// tccr1b bits
#define cs10  v_tccr1a[0]
#define cs11  v_tccr1a[1]
#define cs12  v_tccr1a[2]
#define ctc1  v_tccr1a[3]
#define ices1 v_tccr1a[6]
#define icnc1 v_tccr1a[7]

// tccr2 bits
#define cs20  v_tccr2[0]
#define cs21  v_tccr2[1]
#define cs22  v_tccr2[2]
#define ctc2  v_tccr2[3]
#define com20 v_tccr2[4]
#define com21 v_tccr2[5]
#define pwm2  v_tccr2[6]

// assr bits
// assr = asynchronous status register (for TCNT0)
#define tcr0ub  v_assr[0]
#define ocr0ub  v_assr[1]
#define tcn0ub  v_assr[2]
#define as0     v_assr[3]

// timsk bits
// timer interrupt mask register
#define toie0   v_timsk[0]
#define ocie0   v_timsk[1]
#define toie1   v_timsk[2]
#define ocie1b  v_timsk[3]
#define ocie1a  v_timsk[4]
#define ticie1  v_timsk[5]
#define toie2   v_timsk[6]
#define ocie2   v_timsk[7]

// tifr bits
// timer interrupt flag register
#define tov0    v_tifr[0]
#define ocf0    v_tifr[1]
#define tov1    v_tifr[2]
#define ocf1b   v_tifr[3]
#define ocf1a   v_tifr[4]
#define icf1    v_tifr[5]
#define tov2    v_tifr[6]
#define ocf2    v_tifr[7]

SC_MODULE(timer_counter)
{
  // AVR Control
  sc_in<bool> clk;
  sc_in<bool> reset;

  sc_in<sc_uint<6> > adr;
  sc_in<sc_lv<8> >   dbus_in;
  sc_out<sc_lv<8> >  dbus_out;
  sc_in<sc_logic>    iore;
  sc_in<sc_logic>    iowe;
  sc_out<sc_logic>   out_en;

  // Timer/Counters
  sc_in<sc_logic>  ext1;
  sc_in<sc_logic>  ext2;
  sc_in<sc_logic>  tosc1;
  sc_out<sc_logic> oc0_pwm0;
  sc_out<sc_logic> oc1a_pwm1a;
  sc_out<sc_logic> oc1b_pwm1b;
  sc_out<sc_logic> oc2_pwm2;

  // IRQ
  sc_out<sc_logic> tc0ovfirq;
  sc_in<sc_logic>  tc0ovfirq_ack;
  sc_out<sc_logic> tc0cmpirq;
  sc_in<sc_logic>  tc0cmpirq_ack;
  sc_out<sc_logic> tc2ovfirq;
  sc_in<sc_logic>  tc2ovfirq_ack;
  sc_out<sc_logic> tc2cmpirq;
  sc_in<sc_logic>  tc2cmpirq_ack;
  sc_out<sc_logic> tc1ovfirq;
  sc_in<sc_logic>  tc1ovfirq_ack;
  sc_out<sc_logic> tc1cmpairq;
  sc_in<sc_logic>  tc1cmpairq_ack;
  sc_out<sc_logic> tc1cmpbirq;
  sc_in<sc_logic>  tc1cmpbirq_ack;
  sc_out<sc_logic> tc1icirq;
  sc_in<sc_logic>  tc1icirq_ack;

  // Copies of the external signals
  sc_signal<sc_logic> oc0_pwm0_int;
  sc_signal<sc_logic> oc2_pwm2_int;

  // Registers
  sc_signal<sc_lv<8> > tccr0;
  sc_signal<sc_lv<8> > tccr1a;
  sc_signal<sc_lv<8> > tccr1b;
  sc_signal<sc_lv<8> > tccr2;
  sc_signal<sc_lv<8> > assr;
  sc_signal<sc_lv<8> > timsk;
  sc_signal<sc_lv<8> > tifr;
  sc_signal<sc_lv<8> > tcnt0;
  sc_signal<sc_lv<8> > tcnt2;
  sc_signal<sc_lv<8> > ocr0;
  sc_signal<sc_lv<8> > ocr2;
  sc_signal<sc_lv<8> > tcnt1h;
  sc_signal<sc_lv<8> > tcnt1l;
  sc_signal<sc_lv<8> > ocr1ah;
  sc_signal<sc_lv<8> > ocr1al;
  sc_signal<sc_lv<8> > ocr1bh;
  sc_signal<sc_lv<8> > ocr1bl;
  sc_signal<sc_lv<8> > icr1ah;
  sc_signal<sc_lv<8> > icr1al;
  sc_signal<sc_logic>  tccr0_sel;
  sc_signal<sc_logic>  tccr1a_sel;
  sc_signal<sc_logic>  tccr1b_sel;
  sc_signal<sc_logic>  tccr2_sel;
  sc_signal<sc_logic>  assr_sel;
  sc_signal<sc_logic>  timsk_sel;
  sc_signal<sc_logic>  tifr_sel;
  sc_signal<sc_logic>  tcnt0_sel;
  sc_signal<sc_logic>  tcnt2_sel;
  sc_signal<sc_logic>  ocr0_sel;
  sc_signal<sc_logic>  ocr2_sel;
  sc_signal<sc_logic>  tcnt1h_sel;
  sc_signal<sc_logic>  tcnt1l_sel;
  sc_signal<sc_logic>  ocr1ah_sel;
  sc_signal<sc_logic>  ocr1al_sel;
  sc_signal<sc_logic>  ocr1bh_sel;
  sc_signal<sc_logic>  ocr1bl_sel;
  sc_signal<sc_logic>  icr1ah_sel;
  sc_signal<sc_logic>  icr1al_sel;

  // Prescaler1 signals
  sc_signal<sc_logic> ck8;
  sc_signal<sc_logic> ck64;
  sc_signal<sc_logic> ck256;
  sc_signal<sc_logic> ck1024;
  sc_signal<sc_lv<10> > pre1cnt;
  sc_signal<sc_logic> ext1re;
  sc_signal<sc_logic> ext1fe;
  sc_signal<sc_logic> ext2re;
  sc_signal<sc_logic> ext2fe;

  // Rising/falling edge detectors
  sc_signal<sc_logic> ext1latched;
  sc_signal<sc_logic> ext2latched;

  // Prescalers outputs
  sc_signal<sc_logic> tcnt0_en;
  sc_signal<sc_logic> tcnt0_en2;
  sc_signal<sc_logic> tcnt1_en;
  sc_signal<sc_logic> tcnt2_en;

  // Prescaler0 signals
  sc_signal<sc_logic> pck0;
  sc_signal<sc_logic> pck08;
  sc_signal<sc_logic> pck032;
  sc_signal<sc_logic> pck064;
  sc_signal<sc_logic> pck0128;
  sc_signal<sc_logic> pck0256;
  sc_signal<sc_logic> pck01024;
  sc_signal<sc_logic> tosc1re;
  sc_signal<sc_logic> tosc1latched;
  sc_signal<sc_lv<10> > pre0cnt;
  sc_signal<sc_logic> pck0_del;

  // Timer/counter 0 signals
  sc_signal<sc_lv<8> > tcnt0_tmp;
  sc_signal<sc_lv<8> > tcnt0_in;
  sc_signal<sc_lv<8> > tcnt0_imm_in;
  sc_signal<sc_lv<8> > tccr0_tmp;
  sc_signal<sc_lv<8> > tccr0_in;
  sc_signal<sc_lv<8> > ocr0_tmp;
  sc_signal<sc_lv<8> > ocr0_in;
  sc_signal<sc_logic> tcnt0_cnt_dir;
  sc_signal<sc_logic> tcnt0_clr;
  sc_signal<sc_logic> tcnt0_ld_imm;
  sc_signal<sc_logic> tcnt0_cmp_out;
  sc_signal<sc_logic> tcnt0_inc;

  // For asynchronous mode only
  sc_signal<sc_logic> tcr0ub_tmp;
  sc_signal<sc_logic> ocr0ub_tmp;
  sc_signal<sc_logic> tcn0ub_tmp;

  // Timer/counter 2 signals
  sc_signal<sc_lv<8> > tcnt2_in;
  sc_signal<sc_lv<8> > ocr2_tmp;
  sc_signal<sc_logic> tcnt2_cnt_dir;
  sc_signal<sc_logic> tcnt2_clr;
  sc_signal<sc_lv<8> > tcnt2_imm_in;
  sc_signal<sc_lv<8> > tccr2_tmp;
  sc_signal<sc_lv<8> > ocr2_in;
  sc_signal<sc_logic> tcnt2_ld_imm;
  sc_signal<sc_logic> tcnt2_cmp_out;
  sc_signal<sc_logic> tcnt2_inc;

  // Synchronizer signals
  sc_signal<sc_logic> ext1sa;
  sc_signal<sc_logic> ext1sb;
  sc_signal<sc_logic> ext2sa;
  sc_signal<sc_logic> ext2sb;
  sc_signal<sc_logic> tosc1sa;
  sc_signal<sc_logic> tosc1sb;

  void syncdffs();
  void prescaler_1();
  void tosc1_rising_edge_detector();
  void prescaler_0_cnt();
  void prescaler_0();
  void counter_0();
  void counter_0_inc();
  void tccr0_control();
  void tcnt0_temp_control();
  void tcnt0_temp_cnt();
  void tcnt0_temp_compare();
  void tcnt0_control();
  void tcnt0_compare();
  void output_tcnt0();
  void counter_2();
  void counter_2_inc();
  void tccr2_control();
  void tcnt2_temp_compare();
  void tcnt2_control();
  void tcnt2_compare();
  void output_tcnt2();
  void tifr_bits();
  void timsk_bits();
  void concurrent_statements();
  void set_tcnt0_en();

  SC_CTOR(timer_counter)
  {

    SC_METHOD(syncdffs);
    sensitive_pos << clk;
    SC_METHOD(prescaler_1);
    sensitive_pos << clk;
    SC_METHOD(tosc1_rising_edge_detector);
    sensitive_pos << clk;
    SC_METHOD(prescaler_0_cnt);
    sensitive_pos << clk;
    SC_METHOD(prescaler_0);
    sensitive_pos << clk;
    SC_METHOD(counter_0);
    sensitive_pos << clk;
    SC_METHOD(counter_0_inc);
    sensitive_pos << clk;
    SC_METHOD(tccr0_control);
    sensitive_pos << clk;
    SC_METHOD(tcnt0_temp_control);
    sensitive_pos << clk;
    SC_METHOD(tcnt0_temp_cnt);
    sensitive_pos << clk;
    SC_METHOD(tcnt0_temp_compare);
    sensitive_pos << clk;
    SC_METHOD(tcnt0_control);
    sensitive_pos << clk;
    SC_METHOD(tcnt0_compare);
    sensitive_pos << clk;
    SC_METHOD(output_tcnt0);
    sensitive_pos << clk;
    SC_METHOD(counter_2);
    sensitive_pos << clk;
    SC_METHOD(counter_2_inc);
    sensitive_pos << clk;
    SC_METHOD(tccr2_control);
    sensitive_pos << clk;
    SC_METHOD(tcnt2_temp_compare);
    sensitive_pos << clk;
    SC_METHOD(tcnt2_control);
    sensitive_pos << clk;
    SC_METHOD(tcnt2_compare);
    sensitive_pos << clk;
    SC_METHOD(output_tcnt2);
    sensitive_pos << clk;
    SC_METHOD(tifr_bits);
    sensitive_pos << clk;
    SC_METHOD(timsk_bits);
    sensitive_pos << clk;    

    // Concurrent statements
    SC_METHOD(concurrent_statements);
    sensitive << assr << tccr0 << tccr1a << tccr2 << tcnt0 << tcnt2 << tccr0_tmp;
    sensitive << ocr0 << ocr0_tmp << ocr2 << tifr << timsk;
    sensitive << oc0_pwm0_int << oc2_pwm2_int << iore << iowe << adr;
    sensitive << tccr0_sel << tccr1a_sel << tccr1b_sel << tccr2_sel << assr_sel;
    sensitive << timsk_sel << tifr_sel << tcnt0_sel << tcnt2_sel << ocr0_sel << ocr2_sel;
    sensitive << tcnt1h_sel << tcnt1l_sel << ocr1ah_sel << ocr1al_sel << ocr1bh_sel << ocr1bl_sel;
    sensitive << icr1ah_sel << icr1al_sel;
    sensitive << tosc1re << pck0 << tcn0ub_tmp << tcnt0_ld_imm << tcnt0_inc << tcnt0_cmp_out;
    sensitive << tcnt0_clr << tcnt0_cnt_dir;
    sensitive << dbus_in;
    sensitive << pck08 << pck032 << pck064 << pck0128 << pck0256 << pck01024;

    SC_METHOD(set_tcnt0_en);
    sensitive_pos << clk;
  }
};

#endif
